import asyncio
import itertools
import logging
import math
import random
import webbrowser
from collections import deque
from datetime import datetime, timedelta, timezone

import aiofiles.os
import numpy as np
import pandas as pd
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from PySide6 import QtWidgets

from solie.common import go, outsource
from solie.overlay import DonationGuide, DownloadFillOption
from solie.utility import (
    ApiRequester,
    ApiStreamer,
    DownloadPreset,
    RWLock,
    add_disconnected_functions,
    add_task_duration,
    combine_candle_data,
    download_aggtrade_data,
    fill_holes_with_aggtrades,
    find_stop_flag,
    format_numeric,
    get_current_moment,
    internet_connected,
    make_stop_flag,
    sort_data_frame,
    standardize_candle_data,
)
from solie.widget import overlay
from solie.window import Window

from .united import team

logger = logging.getLogger(__name__)


class Collector:
    def __init__(self, window: Window, scheduler: AsyncIOScheduler):
        # ■■■■■ for data management ■■■■■

        self.window = window
        self.scheduler = scheduler
        self.workerpath = window.datapath / "collector"

        # ■■■■■ internal memory ■■■■■

        self.price_precisions: dict[str, int] = {}  # Symbol and decimal places
        self.markets_gone: list[str] = []  # Symbols

        # ■■■■■ remember and display ■■■■■

        self.api_requester = ApiRequester()

        self.aggtrade_candle_sizes = {}
        for symbol in window.data_settings.target_symbols:
            self.aggtrade_candle_sizes[symbol] = 0

        # Candle data.
        # It's expected to have only the data of current year,
        # while data of previous years are stored in the disk.
        self.candle_data = RWLock(
            standardize_candle_data(window.data_settings.target_symbols)
        )

        # Realtime data chunks
        field_names = itertools.product(
            window.data_settings.target_symbols,
            ("Best Bid Price", "Best Ask Price", "Mark Price"),
        )
        field_names = [str(field_name) for field_name in field_names]
        dtype = [(field_name, np.float32) for field_name in field_names]
        dtpye = [("index", "datetime64[ns]")] + dtype
        self.realtime_data_chunks = RWLock(
            deque([np.recarray(shape=(0,), dtype=dtpye) for _ in range(2)], maxlen=64)
        )

        # Aggregate trades
        field_names = itertools.product(
            window.data_settings.target_symbols,
            ("Price", "Volume"),
        )
        field_names = [str(field_name) for field_name in field_names]
        dtype = [(field_name, np.float32) for field_name in field_names]
        dtpye = [("index", "datetime64[ns]")] + dtype
        self.aggregate_trades = RWLock(np.recarray(shape=(0,), dtype=dtpye))

        # ■■■■■ repetitive schedules ■■■■■

        self.scheduler.add_job(
            self.display_status_information,
            trigger="cron",
            second="*",
        )
        self.scheduler.add_job(
            self.fill_candle_data_holes,
            trigger="cron",
            second="*/10",
        )
        self.scheduler.add_job(
            self.add_candle_data,
            trigger="cron",
            second="*/10",
        )
        self.scheduler.add_job(
            self.organize_data,
            trigger="cron",
            minute="*",
        )
        self.scheduler.add_job(
            self.get_exchange_information,
            trigger="cron",
            minute="*",
        )
        self.scheduler.add_job(
            self.save_candle_data,
            trigger="cron",
            hour="*",
        )

        # ■■■■■ websocket streamings ■■■■■

        self.api_streamers = {
            "MARK_PRICE": ApiStreamer(
                "wss://fstream.binance.com/ws/!markPrice@arr@1s",
                self.add_mark_price,
            ),
        }
        for symbol in self.window.data_settings.target_symbols:
            api_streamer = ApiStreamer(
                f"wss://fstream.binance.com/ws/{symbol.lower()}@bookTicker",
                self.add_book_tickers,
            )
            self.api_streamers[f"BOOK_TICKER_{symbol}"] = api_streamer
            api_streamer = ApiStreamer(
                f"wss://fstream.binance.com/ws/{symbol.lower()}@aggTrade",
                self.add_aggregate_trades,
            )
            self.api_streamers[f"AGG_TRADE_{symbol}"] = api_streamer

        # ■■■■■ invoked by the internet connection status change ■■■■■

        add_disconnected_functions(self.clear_aggregate_trades)

        # ■■■■■ connect UI events ■■■■■

        job = self.guide_donation
        outsource(window.pushButton_9.clicked, job)
        job = self.download_fill_candle_data
        outsource(window.pushButton_2.clicked, job)

        action_menu = QtWidgets.QMenu(self.window)
        self.window.pushButton_13.setMenu(action_menu)

        text = "Open historical data webpage of Binance"
        job = self.open_binance_data_page
        new_action = action_menu.addAction(text)
        outsource(new_action.triggered, job)
        text = "Stop filling candle data"
        job = self.stop_filling_candle_data
        new_action = action_menu.addAction(text)
        outsource(new_action.triggered, job)

    async def load(self):
        await aiofiles.os.makedirs(self.workerpath, exist_ok=True)

        # candle data
        current_year = datetime.now(timezone.utc).year
        async with self.candle_data.write_lock as cell:
            filepath = self.workerpath / f"candle_data_{current_year}.pickle"
            if await aiofiles.os.path.isfile(filepath):
                df: pd.DataFrame = await go(pd.read_pickle, filepath)
                if not df.index.is_monotonic_increasing:
                    df = await go(sort_data_frame, df)
                cell.data = df

    async def organize_data(self):
        start_time = datetime.now(timezone.utc)

        async with self.candle_data.write_lock as cell:
            original_index = cell.data.index
            if not cell.data.index.is_unique:
                unique_index = original_index.drop_duplicates()
                cell.data = cell.data.reindex(unique_index)
            if not cell.data.index.is_monotonic_increasing:
                cell.data = await go(sort_data_frame, cell.data)

        async with self.realtime_data_chunks.write_lock as cell:
            cell.data[-1].sort(order="index")
            if len(cell.data[-1]) > 2**16:
                new_chunk = cell.data[-1][0:0].copy()
                cell.data.append(new_chunk)
                del new_chunk

        async with self.aggregate_trades.write_lock as cell:
            cell.data.sort(order="index")
            last_index = cell.data[-1]["index"]
            slice_from = last_index - np.timedelta64(60, "s")
            mask = cell.data["index"] > slice_from
            cell.data = cell.data[mask].copy()

        duration = (datetime.now(timezone.utc) - start_time).total_seconds()
        add_task_duration("collector_organize_data", duration)

    async def save_candle_data(self):
        # ■■■■■ default values ■■■■■

        current_year = datetime.now(timezone.utc).year
        filepath = self.workerpath / f"candle_data_{current_year}.pickle"
        filepath_new = self.workerpath / f"candle_data_{current_year}.pickle.new"
        filepath_backup = self.workerpath / f"candle_data_{current_year}.pickle.backup"

        async with self.candle_data.read_lock as cell:
            mask = cell.data.index.year == current_year  # type:ignore
            year_df: pd.DataFrame = cell.data[mask].copy()

        # ■■■■■ make a new file ■■■■■

        await go(year_df.to_pickle, filepath_new)

        # ■■■■■ safely replace the existing file ■■■■■

        if await aiofiles.os.path.isfile(filepath_backup):
            await aiofiles.os.remove(filepath_backup)
        if await aiofiles.os.path.isfile(filepath):
            await aiofiles.os.rename(filepath, filepath_backup)
        if await aiofiles.os.path.isfile(filepath_new):
            await aiofiles.os.rename(filepath_new, filepath)

    async def get_exchange_information(self):
        if not internet_connected():
            return

        payload = {}
        response = await self.api_requester.binance(
            http_method="GET",
            path="/fapi/v1/exchangeInfo",
            payload=payload,
        )
        about_exchange = response

        for about_symbol in about_exchange["symbols"]:
            symbol = about_symbol["symbol"]

            about_filter = {}
            for filter in about_symbol["filters"]:
                if filter["filterType"] == "PRICE_FILTER":
                    about_filter = filter
                    break

            ticksize = float(about_filter["tickSize"])
            price_precision = int(math.log10(1 / ticksize))
            self.price_precisions[symbol] = price_precision

    async def fill_candle_data_holes(self):
        # ■■■■■ check internet connection ■■■■■

        if not internet_connected():
            return

        # ■■■■■ moments ■■■■■

        current_moment = get_current_moment()
        split_moment = current_moment - timedelta(days=2)

        # ■■■■■ fill holes ■■■■■

        markets_gone = []
        full_symbols = []
        request_count = 0

        # only the recent part
        async with self.candle_data.read_lock as cell:
            recent_candle_data = cell.data[cell.data.index >= split_moment].copy()

        did_fill = False

        target_symbols = self.window.data_settings.target_symbols
        while len(full_symbols) < len(target_symbols) and request_count < 10:
            for symbol in target_symbols:
                if symbol in full_symbols:
                    continue

                from_moment = current_moment - timedelta(hours=24)
                until_moment = current_moment - timedelta(minutes=1)

                inspect_df: pd.DataFrame = recent_candle_data[symbol][
                    from_moment:until_moment
                ]  # type:ignore
                base_index = inspect_df.dropna().index
                temp_sr = pd.Series(0, index=base_index)
                written_moments = len(temp_sr)

                if written_moments == (86400 - 60) / 10 + 1:
                    # case when there are no holes
                    full_symbols.append(symbol)
                    continue

                if from_moment not in temp_sr.index:
                    temp_sr[from_moment] = np.nan
                if until_moment not in temp_sr.index:
                    temp_sr[until_moment] = np.nan
                temp_sr = await go(temp_sr.asfreq, "10S")
                isnan_sr = await go(temp_sr.isna)
                nan_index = isnan_sr[isnan_sr == 1].index
                moment_to_fill_from: datetime = nan_index[0]  # type:ignore

                # request historical aggtrade data
                aggtrades = {}
                last_fetched_time = moment_to_fill_from
                while last_fetched_time < moment_to_fill_from + timedelta(seconds=10):
                    # intend to fill at least one 10 second candle bar
                    payload = {
                        "symbol": symbol,
                        "startTime": int(last_fetched_time.timestamp() * 1000),
                        "limit": 1000,
                    }
                    response = await self.api_requester.binance(
                        http_method="GET",
                        path="/fapi/v1/aggTrades",
                        payload=payload,
                    )
                    request_count += 1
                    if len(response) == 0:
                        if symbol not in markets_gone:
                            markets_gone.append(symbol)
                        break
                    for aggtrade in response:
                        aggtrade_id = aggtrade["a"]
                        aggtrades[aggtrade_id] = aggtrade
                    last_fetched_id = max(aggtrades.keys())
                    last_fetched_time = datetime.fromtimestamp(
                        aggtrades[last_fetched_id]["T"] / 1000, tz=timezone.utc
                    )

                recent_candle_data = await go(
                    fill_holes_with_aggtrades,
                    symbol,
                    recent_candle_data,
                    aggtrades,
                    moment_to_fill_from,
                    last_fetched_time,
                )
                did_fill = True

        self.markets_gone = markets_gone

        if not did_fill:
            return

        # combine
        async with self.candle_data.write_lock as cell:
            original_candle_data = cell.data[cell.data.index < split_moment]
            # in case the other data is added during the task
            # read the data again
            temp_df = cell.data[cell.data.index >= split_moment]
            recent_candle_data = recent_candle_data.combine_first(temp_df)
            if not recent_candle_data.index.is_monotonic_increasing:
                recent_candle_data = await go(
                    sort_data_frame,
                    recent_candle_data,
                )
            candle_data = pd.concat([original_candle_data, recent_candle_data])
            cell.data = candle_data

    async def display_status_information(self):
        async with self.candle_data.read_lock as cell:
            if len(cell.data) == 0:
                # when the app is executed for the first time
                return

        if len(self.price_precisions) == 0:
            # right after the app execution
            return

        # price
        async with self.aggregate_trades.read_lock as cell:
            ar = cell.data.copy()
        price_precisions = self.price_precisions

        for symbol in self.window.data_settings.target_symbols:
            temp_ar = ar[str((symbol, "Price"))]
            temp_ar = temp_ar[temp_ar != 0]
            if len(temp_ar) > 0:
                price_precision = price_precisions[symbol]
                latest_price = temp_ar[-1]
                text = f"＄{latest_price:.{price_precision}f}"
            else:
                text = "Unavailable"
            self.window.price_labels[symbol].setText(text)

        # bottom information
        if len(self.markets_gone) == 0:
            cumulation_rate = await self.check_candle_data_cumulation_rate()
            async with self.realtime_data_chunks.read_lock as cell:
                chunk_count = len(cell.data)
            first_written_time = None
            last_written_time = None
            for turn in range(chunk_count):
                async with self.realtime_data_chunks.read_lock as cell:
                    if len(cell.data[turn]) > 0:
                        if first_written_time is None:
                            first_record = cell.data[turn][0]
                            first_written_time = first_record["index"]
                            del first_record
                        last_record = cell.data[turn][-1]
                        last_written_time = last_record["index"]
                        del last_record
            if first_written_time is not None and last_written_time is not None:
                written_seconds = last_written_time - first_written_time
                written_seconds = written_seconds.astype(np.int64) / 10**9
            else:
                written_seconds = 0
            written_length = timedelta(seconds=written_seconds)
            range_days = written_length.days
            range_hours, remains = divmod(written_length.seconds, 3600)
            range_minutes, remains = divmod(remains, 60)
            written_length_text = f"{range_days}d {range_hours}h {range_minutes}m"

            text = ""
            text += f"24h candle data accumulation rate {cumulation_rate * 100:.2f}%"
            text += "  ⦁  "
            text += f"Realtime data length {written_length_text}"
        else:
            markets_gone = self.markets_gone
            if len(markets_gone) == 1:
                text = (
                    f"It seems that {markets_gone[0]} market is removed by Binance."
                    + " You should make a new data folder."
                )
            else:
                text = (
                    f"It seems that {', '.join(markets_gone)} markets are removed by Binance."
                    + " You should make a new data folder."
                )

        self.window.label_6.setText(text)

    async def check_candle_data_cumulation_rate(self) -> float:
        current_moment = get_current_moment()
        count_start_moment = current_moment - timedelta(hours=24)
        async with self.candle_data.read_lock as cell:
            cumulated_moments = len(cell.data[count_start_moment:].dropna())
        needed_moments = 24 * 60 * 60 / 10
        cumulation_rate = min(1.0, (cumulated_moments + 2) / needed_moments)
        return cumulation_rate

    async def open_binance_data_page(self):
        await go(webbrowser.open, "https://www.binance.com/en/landing/data")

    async def download_fill_candle_data(self):
        # ■■■■■ ask filling type ■■■■■

        overlay_widget = await overlay(
            "Choose the range to fill",
            DownloadFillOption(),
        )
        filling_type = overlay_widget.result

        if filling_type is None:
            return

        # ■■■■■ prepare target tuples for downloading ■■■■■

        task_id = make_stop_flag("download_fill_candle_data")

        download_presets: list[DownloadPreset] = []
        target_symbols = self.window.data_settings.target_symbols
        if filling_type == 0:
            current_year = datetime.now(timezone.utc).year
            for year in range(2020, current_year):
                for month in range(1, 12 + 1):
                    for symbol in target_symbols:
                        download_presets.append(
                            DownloadPreset(
                                symbol,
                                "monthly",
                                year,
                                month,
                            )
                        )
        elif filling_type == 1:
            current_year = datetime.now(timezone.utc).year
            current_month = datetime.now(timezone.utc).month
            for month in range(1, current_month):
                for symbol in target_symbols:
                    download_presets.append(
                        DownloadPreset(
                            symbol,
                            "monthly",
                            current_year,
                            month,
                        )
                    )
        elif filling_type == 2:
            current_year = datetime.now(timezone.utc).year
            current_month = datetime.now(timezone.utc).month
            current_day = datetime.now(timezone.utc).day
            for target_day in range(1, current_day):
                for symbol in target_symbols:
                    download_presets.append(
                        DownloadPreset(
                            symbol,
                            "daily",
                            current_year,
                            current_month,
                            target_day,
                        )
                    )
        elif filling_type == 3:
            now = datetime.now(timezone.utc)
            yesterday = now - timedelta(hours=24)
            day_before_yesterday = yesterday - timedelta(hours=24)
            for symbol in target_symbols:
                download_presets.append(
                    DownloadPreset(
                        symbol,
                        "daily",
                        day_before_yesterday.year,
                        day_before_yesterday.month,
                        day_before_yesterday.day,
                    ),
                )
                download_presets.append(
                    DownloadPreset(
                        symbol,
                        "daily",
                        yesterday.year,
                        yesterday.month,
                        yesterday.day,
                    ),
                )

        random.shuffle(download_presets)

        total_steps = len(download_presets)
        done_steps = 0

        # ■■■■■ play the progress bar ■■■■■

        async def play_progress_bar():
            while True:
                if find_stop_flag("download_fill_candle_data", task_id):
                    self.window.progressBar_3.setValue(0)
                    return
                else:
                    if done_steps == total_steps:
                        progressbar_value = self.window.progressBar_3.value()
                        if progressbar_value == 1000:
                            await asyncio.sleep(0.1)
                            self.window.progressBar_3.setValue(0)
                            return
                    before_value = self.window.progressBar_3.value()
                    if before_value < 1000:
                        remaining = (
                            math.ceil(1000 / total_steps * done_steps) - before_value
                        )
                        new_value = before_value + math.ceil(remaining * 0.2)
                        self.window.progressBar_3.setValue(new_value)
                    await asyncio.sleep(0.01)

        asyncio.create_task(play_progress_bar())

        # ■■■■■ calculate in parellel ■■■■■

        # Gather information about years.
        current_year = datetime.now(timezone.utc).year
        all_years: set[int] = {t.year for t in download_presets}

        # Download and save historical data by year for lower memory usage.
        # Key is the year, value is the list of download presets.
        classified_download_presets: dict[int, list[DownloadPreset]] = {
            y: [] for y in all_years
        }
        for download_preset in download_presets:
            classified_download_presets[download_preset.year].append(download_preset)

        for preset_year, download_presets in classified_download_presets.items():
            # Make an empty dataframe, but of same types with that of candle data.
            async with self.candle_data.read_lock as cell:
                combined_df = RWLock(cell.data.iloc[0:0].copy())

            async def download_fill(download_preset: DownloadPreset) -> None:
                nonlocal done_steps
                nonlocal combined_df

                if find_stop_flag("download_fill_candle_data", task_id):
                    return

                returned = await go(download_aggtrade_data, download_preset)
                if returned is not None:
                    new_df = returned
                    async with combined_df.write_lock as cell:
                        new = await go(combine_candle_data, new_df, cell.data)
                        cell.data = new

                done_steps += 1

            tasks = [asyncio.create_task(download_fill(p)) for p in download_presets]
            await asyncio.wait(tasks)

            if preset_year < current_year:
                # For data of previous years,
                # save them in the disk.
                async with combined_df.read_lock as cell:
                    await go(
                        cell.data.to_pickle,
                        self.workerpath / f"candle_data_{preset_year}.pickle",
                    )
            else:
                # For data of current year, pass it to this collector worker
                # and store them in the memory.
                async with combined_df.read_lock as cell:
                    async with self.candle_data.write_lock as cell_worker:
                        cell_worker.data = await go(
                            combine_candle_data,
                            cell.data,
                            cell_worker.data,
                        )
                await self.save_candle_data()

        # ■■■■■ add to log ■■■■■

        text = "Filled the candle data with the history data downloaded from Binance"
        logger.info(text)

        # ■■■■■ display to graphs ■■■■■

        asyncio.create_task(team.transactor.display_lines())
        asyncio.create_task(team.simulator.display_lines())
        asyncio.create_task(team.simulator.display_available_years())

    async def add_book_tickers(self, received: dict):
        start_time = datetime.now(timezone.utc)
        symbol = received["s"]
        best_bid = received["b"]
        best_ask = received["a"]
        event_time = np.datetime64(received["E"] * 10**6, "ns")
        async with self.realtime_data_chunks.write_lock as cell:
            original_size = cell.data[-1].shape[0]
            cell.data[-1].resize(original_size + 1, refcheck=False)
            cell.data[-1][-1]["index"] = event_time
            find_key = str((symbol, "Best Bid Price"))
            cell.data[-1][-1][find_key] = best_bid
            find_key = str((symbol, "Best Ask Price"))
            cell.data[-1][-1][find_key] = best_ask
        duration = (datetime.now(timezone.utc) - start_time).total_seconds()
        add_task_duration("add_book_tickers", duration)

    async def add_mark_price(self, received: list):
        start_time = datetime.now(timezone.utc)
        target_symbols = self.window.data_settings.target_symbols
        event_time = np.datetime64(received[0]["E"] * 10**6, "ns")
        filtered_data = {}
        for about_mark_price in received:
            symbol = about_mark_price["s"]
            if symbol in target_symbols:
                mark_price = float(about_mark_price["p"])
                filtered_data[symbol] = mark_price
        async with self.realtime_data_chunks.write_lock as cell:
            original_size = cell.data[-1].shape[0]
            cell.data[-1].resize(original_size + 1, refcheck=False)
            cell.data[-1][-1]["index"] = event_time
            for symbol, mark_price in filtered_data.items():
                find_key = str((symbol, "Mark Price"))
                cell.data[-1][-1][find_key] = mark_price
        duration = (datetime.now(timezone.utc) - start_time).total_seconds()
        add_task_duration("add_mark_price", duration)

    async def add_aggregate_trades(self, received: dict):
        start_time = datetime.now(timezone.utc)
        symbol = received["s"]
        price = float(received["p"])
        volume = float(received["q"])
        trade_time = np.datetime64(received["T"] * 10**6, "ns")
        async with self.aggregate_trades.write_lock as cell:
            original_size = cell.data.shape[0]
            cell.data.resize(original_size + 1, refcheck=False)
            cell.data[-1]["index"] = trade_time
            cell.data[-1][str((symbol, "Price"))] = price
            cell.data[-1][str((symbol, "Volume"))] = volume
        duration = (datetime.now(timezone.utc) - start_time).total_seconds()
        add_task_duration("add_aggregate_trades", duration)

    async def clear_aggregate_trades(self):
        async with self.aggregate_trades.write_lock as cell:
            cell.data = cell.data[0:0].copy()

    async def add_candle_data(self):
        current_moment = get_current_moment()
        before_moment = current_moment - timedelta(seconds=10)

        async with self.aggregate_trades.read_lock as cell:
            data_length = len(cell.data)
        if data_length == 0:
            return

        for _ in range(20):
            async with self.aggregate_trades.read_lock as cell:
                last_received_index = cell.data[-1]["index"]
            if np.datetime64(current_moment) < last_received_index:
                break
            await asyncio.sleep(0.1)

        async with self.aggregate_trades.read_lock as cell:
            aggregate_trades = cell.data.copy()

        first_received_index = aggregate_trades[0]["index"]
        if first_received_index >= np.datetime64(before_moment):
            return

        new_datas = {}

        for symbol in self.window.data_settings.target_symbols:
            block_start_timestamp = before_moment.timestamp()
            block_end_timestamp = current_moment.timestamp()

            index_ar = aggregate_trades["index"].astype(np.int64) / 10**9
            after_start_mask = block_start_timestamp <= index_ar
            before_end_mask = index_ar < block_end_timestamp
            block_ar = aggregate_trades[after_start_mask & before_end_mask]
            block_ar = block_ar[block_ar[str((symbol, "Volume"))] != 0]
            self.aggtrade_candle_sizes[symbol] = block_ar.size

            if len(block_ar) > 0:
                open_price = block_ar[0][str((symbol, "Price"))]
                high_price = block_ar[str((symbol, "Price"))].max()
                low_price = block_ar[str((symbol, "Price"))].min()
                close_price = block_ar[-1][str((symbol, "Price"))]
                sum_volume = block_ar[str((symbol, "Volume"))].sum()
            else:
                async with self.candle_data.read_lock as cell:
                    inspect_sr = cell.data.iloc[-60:][(symbol, "Close")].copy()
                inspect_sr = inspect_sr.dropna()
                if len(inspect_sr) == 0:
                    return
                last_price = inspect_sr.tolist()[-1]
                open_price = last_price
                high_price = last_price
                low_price = last_price
                close_price = last_price
                sum_volume = 0

            new_datas[(symbol, "Open")] = open_price
            new_datas[(symbol, "High")] = high_price
            new_datas[(symbol, "Low")] = low_price
            new_datas[(symbol, "Close")] = close_price
            new_datas[(symbol, "Volume")] = sum_volume

        async with self.candle_data.write_lock as cell:
            for column_name, new_data_value in new_datas.items():
                cell.data.loc[before_moment, column_name] = new_data_value
            if not cell.data.index.is_monotonic_increasing:
                cell.data = await go(sort_data_frame, cell.data)

        duration = (datetime.now(timezone.utc) - current_moment).total_seconds()
        add_task_duration("add_candle_data", duration)

    async def stop_filling_candle_data(self):
        make_stop_flag("download_fill_candle_data")

    async def guide_donation(self):
        await overlay(
            "Support Solie",
            DonationGuide(),
        )

    async def check_saved_years(self) -> list[int]:
        years = [
            int(format_numeric(filename))
            for filename in await aiofiles.os.listdir(self.workerpath)
            if filename.startswith("candle_data_") and filename.endswith(".pickle")
        ]
        return years

    async def read_saved_candle_data(self, year: int) -> pd.DataFrame:
        filepath = self.workerpath / f"candle_data_{year}.pickle"
        candle_data: pd.DataFrame = await go(pd.read_pickle, filepath)
        return candle_data



import re


def format_numeric(string: str) -> str:
    string = re.sub(r"[^0123456789\.]", "", string)
    if string.startswith("."):
        string = "0" + string
    if string.endswith("."):
        string = string[:-1]
    return string



import math

from pyqtgraph import AxisItem


class PercentAxisItem(AxisItem):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def tickValues(self, min_value, max_value, size):  # noqa:N802
        min_value = max(0.001 * max_value, min_value)

        min_power = math.ceil(math.log(min_value, 1.01))
        max_power = math.floor(math.log(max_value, 1.01))

        major_ticks = [1.01**i for i in range(min_power, max_power + 1)]

        minor_ticks = []
        factors = [1 + i * 0.001 for i in range(1, 10)]
        for power in range(min_power - 1, max_power + 1):
            for factor in factors:
                new_tick = factor * 1.01**power
                if min_value < new_tick < max_value:
                    minor_ticks.append(new_tick)

        return [
            (max_value - min_value, major_ticks),
            (max_value - min_value, minor_ticks),
        ]

    def tickStrings(self, tick_values, scale, spacing):  # noqa:N802
        optimal_count = max(2, math.ceil(self.size().height() / 20))
        distance = spacing / optimal_count

        strings = []

        if len(tick_values) > 0:
            next_condition = tick_values[0]
            for tick_value in tick_values:
                if tick_value >= next_condition:
                    next_condition = tick_value + distance
                    string = self.format_fixed_float(tick_value, 6)
                    strings.append(string)
                else:
                    strings.append("")

        return strings

    def format_fixed_float(
        self, number: int | float, width=4, positive_sign=False
    ) -> str:
        if width < 4:
            width = 4

        if number < 0 or (positive_sign and number >= 0):
            # when sign should be included
            absolute_limit = 10 ** (width - 1)
        else:
            absolute_limit = 10**width

        if abs(number) >= absolute_limit:
            number = absolute_limit - 1 if number > 0 else -absolute_limit + 1

        string = f"{number:12.12f}"

        if positive_sign and number >= 0:
            string = "+" + string

        string = string[:width]

        return string



from .coin_selection import CoinSelection
from .datapath_input import DatapathInput
from .donation_guide import DonationGuide
from .download_fill_option import DownloadFillOption
from .long_text_view import LongTextView
from .strategy_basic_input import StrategyBasicInput
from .strategy_develop_input import StrategyDevelopInput
from .strategy_info_view import StrategyInfoView
from .token_selection import TokenSelection

__all__ = [
    "CoinSelection",
    "DatapathInput",
    "DonationGuide",
    "DownloadFillOption",
    "StrategyBasicInput",
    "StrategyDevelopInput",
    "StrategyInfoView",
    "TokenSelection",
    "LongTextView",
]



import io
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from urllib.request import urlopen

import numpy as np
import pandas as pd


@dataclass
class DownloadPreset:
    symbol: str
    unit_size: str  # "daily" or "monthly"
    year: int
    month: int
    day: int = 0  # Valid only when `unit_size` is "daily"


def download_aggtrade_data(download_target: DownloadPreset) -> pd.DataFrame | None:
    symbol = download_target.symbol
    unit_size = download_target.unit_size

    if unit_size == "daily":
        year_string = format(download_target.year, "04")
        month_string = format(download_target.month, "02")
        day_string = format(download_target.day, "02")
        url = (
            "https://data.binance.vision/data/futures/um/daily/aggTrades"
            + f"/{symbol}/{symbol}-aggTrades"
            + f"-{year_string}-{month_string}-{day_string}.zip"
        )
    elif unit_size == "monthly":
        year_string = format(download_target.year, "04")
        month_string = format(download_target.month, "02")
        url = (
            "https://data.binance.vision/data/futures/um/monthly/aggTrades"
            + f"/{symbol}/{symbol}-aggTrades"
            + f"-{year_string}-{month_string}.zip"
        )
    else:
        raise ValueError("This download type is not supported")

    zipped_csv_data = None
    for _ in range(5):
        try:
            zipped_csv_data = io.BytesIO(urlopen(url).read())
            break
        except Exception:
            pass

    if zipped_csv_data is None:
        return

    try:
        df: pd.DataFrame = pd.concat(
            pd.read_csv(
                zipped_csv_data,
                compression="zip",
                header=None,
                usecols=[1, 2, 5],
                dtype={1: np.float32, 2: np.float32, 5: np.int64},
                chunksize=10**6,
            )
        )
    except ValueError:
        # when there is a header line in start of the file
        # from august 2022, header is included from binance
        df: pd.DataFrame = pd.concat(
            pd.read_csv(
                zipped_csv_data,
                compression="zip",
                header=0,
                usecols=[1, 2, 5],
                dtype={1: np.float32, 2: np.float32, 5: np.int64},
                chunksize=10**6,
            )
        )
        df = df.rename(columns={"price": 1, "quantity": 2, "transact_time": 5})

    df = df.set_index(5)
    df.index.name = None
    df.index = pd.to_datetime(df.index, unit="ms", utc=True)

    # fill index holes that's smaller than 10 minutes
    temp_sr = pd.Series(0, index=df.index, dtype=np.float32)
    temp_sr = temp_sr.groupby(temp_sr.index).first()
    temp_sr = temp_sr.resample("10s").agg("mean")
    temp_sr = temp_sr.interpolate(limit=60, limit_direction="forward")
    temp_sr = temp_sr.replace(np.nan, 1)
    temp_sr = temp_sr.replace(0, np.nan)
    temp_sr = temp_sr.interpolate(limit=60, limit_direction="backward")
    temp_sr = temp_sr.replace(np.nan, 0)
    temp_sr = temp_sr.replace(1, np.nan)
    temp_sr = temp_sr.dropna()
    valid_index = temp_sr.index

    del temp_sr

    # process data

    close_sr: pd.Series = df[1].resample("10s").agg("last")  # type: ignore
    close_sr = close_sr.reindex(valid_index)
    close_sr = close_sr.ffill()
    close_sr = close_sr.astype(np.float32)
    close_sr.name = (symbol, "Close")

    open_sr: pd.Series = df[1].resample("10s").agg("first")  # type: ignore
    open_sr = open_sr.reindex(valid_index)
    open_sr = open_sr.fillna(value=close_sr)
    open_sr = open_sr.astype(np.float32)
    open_sr.name = (symbol, "Open")

    high_sr: pd.Series = df[1].resample("10s").agg("max")  # type: ignore
    high_sr = high_sr.reindex(valid_index)
    high_sr = high_sr.fillna(value=close_sr)
    high_sr = high_sr.astype(np.float32)
    high_sr.name = (symbol, "High")

    low_sr: pd.Series = df[1].resample("10s").agg("min")  # type: ignore
    low_sr = low_sr.reindex(valid_index)
    low_sr = low_sr.fillna(value=close_sr)
    low_sr = low_sr.astype(np.float32)
    low_sr.name = (symbol, "Low")

    volume_sr: pd.Series = df[2].resample("10s").agg("sum")  # type: ignore
    volume_sr = volume_sr.reindex(valid_index)
    volume_sr = volume_sr.fillna(value=0)
    volume_sr = volume_sr.astype(np.float32)
    volume_sr.name = (symbol, "Volume")

    del df

    series_list = [open_sr, high_sr, low_sr, close_sr, volume_sr]
    new_df = pd.concat(series_list, axis="columns")

    return new_df


def fill_holes_with_aggtrades(
    symbol: str,
    recent_candle_data: pd.DataFrame,
    aggtrades: pd.DataFrame,
    moment_to_fill_from: datetime,
    last_fetched_time: datetime,
) -> pd.DataFrame:
    fill_moment = moment_to_fill_from

    while fill_moment < last_fetched_time - timedelta(seconds=10):
        block_start = fill_moment
        block_end = fill_moment + timedelta(seconds=10)

        aggtrade_prices = []
        aggtrade_volumes = []
        for _, aggtrade in sorted(aggtrades.items()):
            # sorted by time
            aggtrade_time = datetime.fromtimestamp(
                aggtrade["T"] / 1000, tz=timezone.utc
            )
            if block_start <= aggtrade_time < block_end:
                aggtrade_prices.append(float(aggtrade["p"]))
                aggtrade_volumes.append(float(aggtrade["q"]))

        can_write = True

        if len(aggtrade_prices) == 0:
            # when there are no trades
            inspect_sr = recent_candle_data[(symbol, "Close")]
            inspect_sr = inspect_sr.sort_index()
            last_prices = inspect_sr[:fill_moment].dropna()
            if len(last_prices) == 0:
                # when there are no previous data
                # because new data folder was created
                can_write = False
                last_price = 0
            else:
                last_price = last_prices.iloc[-1]
            open_price = last_price
            high_price = last_price
            low_price = last_price
            close_price = last_price
            sum_volume = 0
        else:
            open_price = aggtrade_prices[0]
            high_price = max(aggtrade_prices)
            low_price = min(aggtrade_prices)
            close_price = aggtrade_prices[-1]
            sum_volume = sum(aggtrade_volumes)

        if can_write:
            column = (symbol, "Open")
            recent_candle_data.loc[fill_moment, column] = open_price
            column = (symbol, "High")
            recent_candle_data.loc[fill_moment, column] = high_price
            column = (symbol, "Low")
            recent_candle_data.loc[fill_moment, column] = low_price
            column = (symbol, "Close")
            recent_candle_data.loc[fill_moment, column] = close_price
            column = (symbol, "Volume")
            recent_candle_data.loc[fill_moment, column] = sum_volume

        fill_moment += timedelta(seconds=10)

    recent_candle_data = recent_candle_data.sort_index(axis="index")
    recent_candle_data = recent_candle_data.sort_index(axis="columns")

    return recent_candle_data



# https://stackoverflow.com/questions/67029993/pyqt-creating-a-popup-in-the-window
import asyncio
from typing import TypeVar

from PySide6 import QtCore, QtGui, QtWidgets

from solie.common import outsource

from .popup_box import PopupBox
from .transparent_scroll_area import TransparentScrollArea


class BaseOverlay(QtWidgets.QWidget):
    done_event = asyncio.Event()

    async def confirm_closing(self) -> bool:
        return True


W = TypeVar("W", bound=BaseOverlay)


# show an mainpulatable overlap popup
async def overlay(title: str, widget: W, close_button=True) -> W:
    overlay_panel = OverlayPopup(title, widget, close_button)
    overlay_panel.show()

    await widget.done_event.wait()
    overlay_panel.setParent(None)

    return widget


class OverlayPopup(QtWidgets.QWidget):
    installed_window: QtWidgets.QMainWindow

    def showEvent(self, event):  # noqa:N802
        # needed for filling the window when resized
        parent: QtWidgets.QMainWindow = self.parent()  # type:ignore
        self.setGeometry(parent.rect())

    def eventFilter(self, source, event):  # noqa:N802
        # needed for filling the window when resized
        if event.type() == event.Type.Resize:
            self.setGeometry(source.rect())
        return super().eventFilter(source, event)

    @classmethod
    def install_window(cls, window: QtWidgets.QMainWindow):
        cls.installed_window = window

    def __init__(
        self,
        title: str,
        widget: BaseOverlay,
        close_button: bool,
    ):
        # ■■■■■ the basic ■■■■■

        super().__init__(self.installed_window)

        # ■■■■■ set properties ■■■■■

        # needed for filling the window when resized
        self.installed_window.installEventFilter(self)

        # ■■■■■ in case other overlay popup exists ■■■■■

        widget.done_event.set()
        widget.done_event.clear()

        # ■■■■■ prepare answer ■■■■■

        self.answer = 0

        # ■■■■■ full structure ■■■■■

        self.setAttribute(QtCore.Qt.WidgetAttribute.WA_DeleteOnClose)
        self.setAttribute(QtCore.Qt.WidgetAttribute.WA_StyledBackground)
        self.setAutoFillBackground(True)

        # ■■■■■ full layout ■■■■■

        full_layout = QtWidgets.QHBoxLayout(self)

        # ■■■■■ visaul box ■■■■■

        # box
        content_box = PopupBox()
        content_box.setAutoFillBackground(True)
        content_box.setMaximumSize(1600, 1200)
        content_box_layout = QtWidgets.QVBoxLayout(content_box)
        full_layout.addWidget(content_box)

        # line containing the title and close button
        this_layout = QtWidgets.QHBoxLayout()
        title_label = QtWidgets.QLabel(title)
        title_label_font = QtGui.QFont()
        title_label_font.setPointSize(12)
        title_label.setFont(title_label_font)
        title_label.setWordWrap(False)
        this_layout.addWidget(title_label)
        top_widget = QtWidgets.QSpacerItem(
            0,
            0,
            QtWidgets.QSizePolicy.Policy.Expanding,
            QtWidgets.QSizePolicy.Policy.Minimum,
        )
        this_layout.addItem(top_widget)
        if close_button:
            close_button_widget = QtWidgets.QPushButton("✕", content_box)
            close_button_font = QtGui.QFont()
            close_button_font.setPointSize(11)
            close_button_widget.setFont(close_button_font)

            async def job():
                should_close = await widget.confirm_closing()
                if should_close:
                    widget.done_event.set()

            outsource(close_button_widget.clicked, job)
            this_layout.addWidget(close_button_widget)
        content_box_layout.addLayout(this_layout)

        # scroll area
        scroll_area = TransparentScrollArea()

        scroll_area.setWidget(widget)
        scroll_area.setWidgetResizable(True)

        content_box_layout.addWidget(scroll_area)



from PySide6 import QtCore, QtGui, QtWidgets

from solie.common import outsource

from .overlay_popup import BaseOverlay, overlay


class LogOverlay(BaseOverlay):
    def __init__(self, log_content: str):
        # ■■■■■ the basic ■■■■■

        super().__init__()

        # ■■■■■ full layout ■■■■■

        full_layout = QtWidgets.QVBoxLayout(self)
        cards_layout = QtWidgets.QVBoxLayout()
        full_layout.addLayout(cards_layout)

        label = QtWidgets.QLabel(log_content)
        fixed_width_font = QtGui.QFont("Source Code Pro", 9)
        label.setFont(fixed_width_font)
        label.setTextInteractionFlags(
            QtCore.Qt.TextInteractionFlag.TextSelectableByMouse
        )
        cards_layout.addWidget(label)


class LogList(QtWidgets.QListWidget):
    def __init__(self, parent):
        super().__init__(parent)
        self.fixed_width_font = QtGui.QFont("Source Code Pro", 9)
        self.setFont(self.fixed_width_font)
        outsource(self.itemClicked, self.show_fulltext)

    def add_item(self, summarization: str, log_content: str):
        maximum_item_limit = 1024

        new_item = QtWidgets.QListWidgetItem(self)
        new_item.setText(summarization)
        new_item.setData(QtCore.Qt.ItemDataRole.UserRole, log_content)

        super().addItem(new_item)

        index_count = self.count()

        if index_count > maximum_item_limit:
            remove_count = index_count - maximum_item_limit
            for _ in range(remove_count):
                self.takeItem(0)

    async def show_fulltext(self):
        selected_index = self.currentRow()

        selected_item = self.item(selected_index)
        text = selected_item.data(QtCore.Qt.ItemDataRole.UserRole)

        await overlay(
            "This is the full log",
            LogOverlay(text),
        )



import asyncio
import json
import logging
from typing import Callable, Coroutine

import aiohttp

logger = logging.getLogger(__name__)


class ApiStreamError(Exception):
    def __init__(self, received: dict | list):
        formatted = json.dumps(received, indent=2)
        super().__init__(formatted)


class ApiStreamer:
    def __init__(
        self,
        url: str,
        handler: Callable[[dict], Coroutine] | Callable[[list], Coroutine],
    ):
        self._url = url
        self._handler = handler
        self.session = aiohttp.ClientSession()

        if url != "":
            asyncio.create_task(self._run_websocket())

    def __del__(self):
        asyncio.create_task(self._close_self())

    async def _run_websocket(self):
        while True:
            try:
                async with self.session.ws_connect(self._url) as websocket:
                    logger.info(f"Websocket connected: {self._url}")
                    async for received_raw in websocket:
                        if received_raw.type in (
                            aiohttp.WSMsgType.CLOSED,
                            aiohttp.WSMsgType.ERROR,
                        ):
                            logger.info(f"Websocket closed: {self._url}")
                            break
                        else:
                            received = received_raw.json()

                            def done_callback(task: asyncio.Task, received=received):
                                error = task.exception()
                                if error:
                                    raise ApiStreamError(received) from error

                            task = asyncio.create_task(self._handler(received))
                            task.add_done_callback(done_callback)
                    logger.info(f"Websocket stopped: {self._url}")
            except Exception as error:
                # Handle errors that might occur due to network issues
                logger.exception(f"Websocket error: {error}")
                # Wait for a few seconds before attempting to reconnect
                await asyncio.sleep(5)

    async def _close_self(self):
        await self.session.close()



import asyncio
from dataclasses import replace

import aiofiles
import aiofiles.os
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from PySide6 import QtGui, QtWidgets

from solie.common import PACKAGE_PATH, outsource
from solie.overlay import StrategyBasicInput, StrategyDevelopInput
from solie.utility import Strategies, Strategy, create_strategy_code_name
from solie.widget import ask, overlay
from solie.window import Window


class Strategiest:
    def __init__(self, window: Window, scheduler: AsyncIOScheduler):
        # ■■■■■ for data management ■■■■■

        self.window = window
        self.scheduler = scheduler
        self.workerpath = window.datapath / "strategist"

        # ■■■■■ internal memory ■■■■■

        # ■■■■■ remember and display ■■■■■

        self.strategies: Strategies
        self.strategy_cards: list[QtWidgets.QGroupBox] = []

        self.before_selections = {}

        iconpath = PACKAGE_PATH / "static" / "icon"
        self.red_pixmap = QtGui.QPixmap()
        self.red_pixmap.load(str(iconpath / "traffic_light_red.png"))
        self.yellow_pixmap = QtGui.QPixmap()
        self.yellow_pixmap.load(str(iconpath / "traffic_light_yellow.png"))
        self.green_pixmap = QtGui.QPixmap()
        self.green_pixmap.load(str(iconpath / "traffic_light_green.png"))

        # ■■■■■ repetitive schedules ■■■■■

        # ■■■■■ websocket streamings ■■■■■

        self.api_streamers = {}

        # ■■■■■ invoked by the internet connection status change ■■■■■

        # ■■■■■ connect UI events ■■■■■

        job = self.add_blank_strategy
        outsource(window.pushButton_5.clicked, job)

    async def load(self):
        await aiofiles.os.makedirs(self.workerpath, exist_ok=True)

        filepath = self.workerpath / "strategies.json"
        if await aiofiles.os.path.isfile(filepath):
            async with aiofiles.open(filepath, "r", encoding="utf8") as file:
                self.strategies = Strategies.from_json(await file.read())
        else:
            first_strategy = Strategy(
                code_name="SLIESS",
                readable_name="Sample Strategy",
                description="Not for real investment."
                + " This strategy is only for demonstration purposes.",
            )
            filepath = PACKAGE_PATH / "static" / "sample_indicators_script.txt"
            async with aiofiles.open(filepath, "r", encoding="utf8") as file:
                read_data = await file.read()
                first_strategy.indicators_script = read_data
            filepath = PACKAGE_PATH / "static" / "sample_decision_script.txt"
            async with aiofiles.open(filepath, "r", encoding="utf8") as file:
                read_data = await file.read()
                first_strategy.decision_script = read_data
            self.strategies = Strategies(all=[first_strategy])

    async def save_strategies(self):
        filepath = self.workerpath / "strategies.json"
        async with aiofiles.open(filepath, "w", encoding="utf8") as file:
            await file.write(self.strategies.to_json(indent=2))

    async def display_strategies(self):
        self.window.comboBox_2.clear()
        self.window.comboBox.clear()
        for strategy_card in self.strategy_cards:
            strategy_card.setParent(None)
        self.strategy_cards = []

        for strategy in self.strategies.all:
            if strategy.risk_level == 2:
                icon_pixmap = self.red_pixmap
            elif strategy.risk_level == 1:
                icon_pixmap = self.yellow_pixmap
            elif strategy.risk_level == 0:
                icon_pixmap = self.green_pixmap
            else:
                raise ValueError("Invalid risk level for drawing an icon")
            text = f"{strategy.code_name} {strategy.version} - {strategy.readable_name}"
            traffic_light_icon = QtGui.QIcon()
            traffic_light_icon.addPixmap(icon_pixmap)

            self.window.comboBox_2.addItem(traffic_light_icon, text)
            self.window.comboBox.addItem(traffic_light_icon, text)

            strategy_card = QtWidgets.QGroupBox()
            self.window.verticalLayout_16.addWidget(strategy_card)
            self.strategy_cards.append(strategy_card)
            card_layout = QtWidgets.QHBoxLayout(strategy_card)

            icon_label = QtWidgets.QLabel("")
            card_layout.addWidget(icon_label)
            icon_label.setPixmap(icon_pixmap)
            icon_label.setScaledContents(True)
            icon_label.setFixedSize(16, 16)

            text_label = QtWidgets.QLabel(text)
            card_layout.addWidget(text_label)

            spacer = QtWidgets.QSpacerItem(
                0,
                0,
                QtWidgets.QSizePolicy.Policy.Expanding,
                QtWidgets.QSizePolicy.Policy.Minimum,
            )
            card_layout.addItem(spacer)

            async def job_bs(strategy=strategy):
                await self.remember_strategy_selections()
                await overlay(
                    "Develop your strategy",
                    StrategyDevelopInput(strategy),
                )
                asyncio.create_task(self.display_strategies())
                asyncio.create_task(self.save_strategies())
                asyncio.create_task(self.restore_strategy_selections())

            edit_button = QtWidgets.QPushButton("Develop")
            card_layout.addWidget(edit_button)
            outsource(edit_button.clicked, job_bs)

            async def job_eb(strategy=strategy):
                await self.remember_strategy_selections()
                await overlay(
                    "Edit your strategy's basic information",
                    StrategyBasicInput(strategy),
                )
                asyncio.create_task(self.display_strategies())
                asyncio.create_task(self.save_strategies())
                asyncio.create_task(self.restore_strategy_selections())

            edit_button = QtWidgets.QPushButton("Edit basic info")
            card_layout.addWidget(edit_button)
            outsource(edit_button.clicked, job_eb)

            action_menu = QtWidgets.QMenu(self.window)
            action_button = QtWidgets.QPushButton()
            action_button.setText("☰")
            action_button.setMenu(action_menu)
            card_layout.addWidget(action_button)

            async def job_rs(strategy=strategy):
                answer = await ask(
                    "Remove this strategy?",
                    "Once you remove this, it cannot be recovered.",
                    ["Remove"],
                )
                if answer == 0:
                    return
                await self.remember_strategy_selections()
                self.strategies.all.remove(strategy)
                asyncio.create_task(self.display_strategies())
                asyncio.create_task(self.save_strategies())
                asyncio.create_task(self.restore_strategy_selections())

            new_action = action_menu.addAction("Remove")
            outsource(new_action.triggered, job_rs)

            async def job_dp(strategy=strategy):
                await self.remember_strategy_selections()
                duplicated = replace(
                    strategy,
                    code_name=create_strategy_code_name(),
                )
                self.strategies.all.append(duplicated)
                asyncio.create_task(self.display_strategies())
                asyncio.create_task(self.save_strategies())
                asyncio.create_task(self.restore_strategy_selections())

            new_action = action_menu.addAction("Duplicate")
            outsource(new_action.triggered, job_dp)

            async def job_ss(strategy=strategy):
                await self.remember_strategy_selections()
                original_index = self.strategies.all.index(strategy)
                after_index = original_index + 1
                self.strategies.all.pop(original_index)
                self.strategies.all.insert(after_index, strategy)
                asyncio.create_task(self.display_strategies())
                asyncio.create_task(self.save_strategies())
                asyncio.create_task(self.restore_strategy_selections())

            edit_button = QtWidgets.QPushButton("▼")
            card_layout.addWidget(edit_button)
            outsource(edit_button.clicked, job_ss)

            async def job_us(strategy=strategy):
                await self.remember_strategy_selections()
                original_index = self.strategies.all.index(strategy)
                after_index = original_index - 1
                self.strategies.all.pop(original_index)
                self.strategies.all.insert(after_index, strategy)
                asyncio.create_task(self.display_strategies())
                asyncio.create_task(self.save_strategies())
                asyncio.create_task(self.restore_strategy_selections())

            edit_button = QtWidgets.QPushButton("▲")
            card_layout.addWidget(edit_button)
            outsource(edit_button.clicked, job_us)

    async def add_blank_strategy(self):
        await self.remember_strategy_selections()
        new_strategy = Strategy(code_name=create_strategy_code_name())
        self.strategies.all.append(new_strategy)
        await self.display_strategies()
        await self.restore_strategy_selections()

    async def remember_strategy_selections(self):
        before_index = self.window.comboBox_2.currentIndex()
        self.before_selections["transactor"] = self.strategies.all[before_index]
        before_index = self.window.comboBox.currentIndex()
        self.before_selections["simulator"] = self.strategies.all[before_index]

    async def restore_strategy_selections(self):
        if self.before_selections["transactor"] in self.strategies.all:
            new_index = self.strategies.all.index(self.before_selections["transactor"])
            self.window.comboBox_2.setCurrentIndex(new_index)
        if self.before_selections["simulator"] in self.strategies.all:
            new_index = self.strategies.all.index(self.before_selections["simulator"])
            self.window.comboBox.setCurrentIndex(new_index)



from .analyze_market import (
    CalculationInput,
    CalculationOutput,
    SimulationError,
    decide,
    make_indicators,
    simulate_chunk,
)
from .api_requester import ApiRequester, ApiRequestError
from .api_streamer import ApiStreamer
from .backward_compatibility import examine_data_files
from .ball import ball_ceil, ball_floor
from .check_internet import (
    add_connected_functions,
    add_disconnected_functions,
    internet_connected,
    is_internet_checked,
    monitor_internet,
)
from .compare_versions import is_left_version_higher
from .convert import list_to_dict, value_to_indexes
from .download_from_binance import (
    DownloadPreset,
    download_aggtrade_data,
    fill_holes_with_aggtrades,
)
from .log_handler import LogHandler
from .pandas_related import combine_candle_data
from .percent_axis_item import PercentAxisItem
from .rw_lock import RWLock
from .simply_format import format_numeric
from .sort_pandas import sort_data_frame, sort_series
from .standardize import (
    create_strategy_code_name,
    standardize_account_state,
    standardize_asset_record,
    standardize_candle_data,
    standardize_unrealized_changes,
)
from .stop_flag import find_stop_flag, make_stop_flag
from .structs import (
    SimulationSettings,
    SimulationSummary,
    Strategies,
    Strategy,
    TransactionSettings,
)
from .syntax_highlighter import SyntaxHighlighter
from .time_axis_item import TimeAxisItem
from .timing import add_task_duration, get_current_moment, get_task_duration
from .user_settings import (
    DataSettings,
    read_data_settings,
    read_datapath,
    save_data_settings,
    save_datapath,
)

__all__ = [
    "ApiRequester",
    "ApiRequestError",
    "ApiStreamer",
    "ball_ceil",
    "ball_floor",
    "add_connected_functions",
    "add_disconnected_functions",
    "internet_connected",
    "monitor_internet",
    "combine_candle_data",
    "is_left_version_higher",
    "list_to_dict",
    "value_to_indexes",
    "decide",
    "download_aggtrade_data",
    "examine_data_files",
    "fill_holes_with_aggtrades",
    "LogHandler",
    "make_indicators",
    "PercentAxisItem",
    "add_task_duration",
    "get_task_duration",
    "RWLock",
    "format_numeric",
    "sort_data_frame",
    "sort_series",
    "standardize_candle_data",
    "standardize_account_state",
    "standardize_asset_record",
    "standardize_unrealized_changes",
    "create_strategy_code_name",
    "make_stop_flag",
    "find_stop_flag",
    "Strategy",
    "Strategies",
    "TransactionSettings",
    "SimulationSettings",
    "SimulationSummary",
    "SyntaxHighlighter",
    "TimeAxisItem",
    "read_data_settings",
    "read_datapath",
    "save_data_settings",
    "save_datapath",
    "DataSettings",
    "DownloadPreset",
    "CalculationInput",
    "CalculationOutput",
    "simulate_chunk",
    "SimulationError",
    "is_internet_checked",
    "monitor_internet",
    "get_current_moment",
]



from dataclasses import dataclass

from dataclasses_json import DataClassJsonMixin


@dataclass
class Strategy(DataClassJsonMixin):
    code_name: str
    readable_name: str = "A New Blank Strategy"
    version: str = "1.0"
    description: str = "A blank strategy template before being written"
    risk_level: int = 2  # 2 means high, 1 means middle, 0 means low
    parallelized_simulation: bool = False
    chunk_division: int = 30
    indicators_script: str = "pass"
    decision_script: str = "pass"


@dataclass
class Strategies(DataClassJsonMixin):
    all: list[Strategy]


@dataclass
class TransactionSettings(DataClassJsonMixin):
    strategy_index: int = 0
    should_transact: bool = False
    desired_leverage: int = 1
    binance_api_key: str = ""
    binance_api_secret: str = ""


@dataclass
class SimulationSettings:
    year: int
    strategy_index: int = 0
    maker_fee: float = 0.02
    taker_fee: float = 0.04
    leverage: int = 1


@dataclass
class SimulationSummary:
    year: int
    strategy_code_name: str
    strategy_version: str



from PySide6 import QtWidgets


class Gauge(QtWidgets.QPushButton):
    def __init__(self, parent):
        super().__init__(parent)
        self.is_mouse_over = False
        self.normal_text = ""
        self.lock_text = "🔒"

    def enterEvent(self, event):  # noqa:N802
        self.is_mouse_over = True
        self.setFlat(False)
        super().setText(self.lock_text)
        super().enterEvent(event)

    def leaveEvent(self, event):  # noqa:N802
        self.is_mouse_over = False
        self.setFlat(True)
        super().setText(self.normal_text)
        super().leaveEvent(event)

    def setText(self, text):  # noqa:N802
        self.normal_text = text
        if self.is_mouse_over:
            super().setText(self.lock_text)
        else:
            super().setText(self.normal_text)
